#include "global.h"
#include "util.h"

using namespace std;

///////////////////////////////////////////////////////////////////////////////
//
//  Morphing Algorithms follow. Your code here.
//
///////////////////////////////////////////////////////////////////////////////
/*
Morphing consists of distorting each image, and blending the two results. 
I suggest you first write a blend function that generates each output pixel 
as a linear interpolation of the two input pixels at the same location.
Once you have this working, you can blend the two input images and see
whether things are working correctly. This should produce an image that looks
like both input images are semi-transparent and overlaid on one another.

Once you have blend working, you can attempt the much harder distortion function.
This takes a list of features vectors at the start and end of the time, and an image
corresponding to the image at the start. It then interpolates each feature vector 
as a combination of the start and end feature vector. This new list of feature vectors
is used to distort the image according to the algorithm described in
Feature-Based Image Metamorphosos, T. Beier, 1992

You will use the distort function on both input images to create a distorted version
of both images at the same point in time, which means that img1 needs to be
moved from 0 to t (total amount t), while img2 needs to be moved from 1 to t (total amunt 1-t).
Since we want to write a single function to do this, you need to flip the features and the t value
you pass to the distort function for img2 versus img1. 

A good way to debug the distory function is to return the result of the distort on img1 or img2
without blending the two, so you can see the result distort works. Good luck!

*/


#define DEBUG true

void lerpFeatures(vector<FeatureVector> featuresStart, 
	                   vector<FeatureVector> featuresEnd,
	                   vector<FeatureVector>& featuresResult,
	                   double t) {
	for (unsigned int i = 0; i < featuresStart.size(); i++) {
		FeatureVector res;
		res.P[0] = (1-t)*featuresStart[i].P[0] + (t)*featuresEnd[i].P[0];
		res.P[1] = (1-t)*featuresStart[i].P[1] + (t)*featuresEnd[i].P[1];
		res.Q[0] =   (1-t)*featuresStart[i].Q[0]   + (t)*featuresEnd[i].Q[0];
		res.Q[1] =   (1-t)*featuresStart[i].Q[1]   + (t)*featuresEnd[i].Q[1];
		featuresResult.push_back(res);
	}
}

inline vec2 perpendicular(vec2 x) {
	return vec2(-x[1], x[0]);
}

inline double getU(FeatureVector dest, vec2 X) {
	return ((X - dest.P) * (dest.Q - dest.P)) / ((dest.Q - dest.P).length2());
}
inline double getV(FeatureVector dest, vec2 X) {
	return ((X - dest.P) * perpendicular(dest.Q - dest.P))  / ((dest.Q - dest.P).length());
}

inline vec2 getDX(FeatureVector orig, double u, double v) {
	return orig.P + u*(orig.Q - orig.P) + (v*perpendicular(orig.Q - orig.P))/((orig.Q - orig.P).length());
}
double getShortestDistanceToLine(vec2 P, vec2 Q, vec2 X) {
	double qp2 = (Q - P).length2();
	double dist = (((Q - P) * (X - P)) / qp2);
	if (dist < 0)
		dist = 0;
	if (dist > 1)
		dist = 1;
	
	vec2 cP = (P + (Q-P)*dist);
	return (X - cP).length();
}

double getShortestDistanceToFeatureVector(FeatureVector f, double u, double v, vec2 X) {
	double dist = 0;
	if (u < 0) {
		dist = (f.P - X).length();
	} else if (u > 1) {
		dist = (f.Q - X).length();
	} else {
		dist = abs((int)v);
	}
	return dist;
}

void bilerpColor(FIBITMAP* image, vec2 X, RGBQUAD* c) {
	int Dx = (int)X[0];
	int Dy = (int)X[1];
	if (Dx < 0)
		Dx = 0;
	if (Dy < 0)
		Dy = 0;
	if (Dx >= (int)FreeImage_GetWidth(image))
		Dx = FreeImage_GetWidth(image) - 1;
	if (Dy >= (int)FreeImage_GetHeight(image))
		Dy = FreeImage_GetHeight(image) - 1;
	FreeImage_GetPixelColor(image, Dx, Dy, c);
}
/*
Distorts an image according to the algorithm described in Feature-Based Image Metamorphosis.
Linearly interpolates the features vectors from featuresStart to featuresEnd
*/
FIBITMAP* distortImage(vector<FeatureVector> featuresStart, 
	                   vector<FeatureVector> featuresEnd,
	                   FIBITMAP* image,
	                   double t,
	                   double a, double b, double p) {
	
	vector<FeatureVector> featuresLerp;
	lerpFeatures(featuresStart, featuresEnd, featuresLerp, t);
	
	int width = FreeImage_GetWidth(image);
	int height = FreeImage_GetHeight(image);
	int bpp = FreeImage_GetBPP(image);
	FIBITMAP *output = FreeImage_Allocate(width, height, bpp);
	
	if (!output) {
		cerr << "Could not allocate bitmap!" << endl;
		exit(1);
	}
	
	cout << "Distorting image..." << endl;
	
	RGBQUAD c;
	for (int x = 0; x < width; x++) {
		for (int y = 0; y < height; y++) {
			vec2 X = vec2((double)x,(double)y);
			vec2 DSUM(0.0,0.0);
			double weightsum = 0;
			
			
			 for (unsigned int vi = 0; vi < featuresLerp.size(); vi++) {
				double u = getU(featuresLerp[vi], X);
				double v = getV(featuresLerp[vi], X);
				vec2 Xprime = getDX(featuresStart[vi], u, v);
				vec2 D = Xprime - X;
				
				//double dist = getShortestDistanceToLine(featuresLerp[vi].P, featuresLerp[vi].Q, X);
				double dist = getShortestDistanceToFeatureVector(featuresLerp[vi], u, v, X);
				
				double lengthp = pow((featuresLerp[vi].Q - featuresLerp[vi].P).length(), p);
				double weight = pow(lengthp / (a + dist), b);
			//	double weight = 1;
				DSUM += (D * weight);
				weightsum += weight;
			}
			
			vec2 Xprime = X + (DSUM / weightsum);
			
			bilerpColor(image, Xprime, &c);
			FreeImage_SetPixelColor(output, x, y, &c);
		}
	}
	return output;
}

/*
Linearly blends each pixel of two images to produce the resulting image.
*/
FIBITMAP* blendImages(FIBITMAP *img1, 
	                  FIBITMAP *img2, 
	                  double t) {
		
	int width = FreeImage_GetWidth(img1);
	int height = FreeImage_GetHeight(img1);
	int bpp = FreeImage_GetBPP(img1);
	FIBITMAP *output = FreeImage_Allocate(width, height, bpp);
	
	if (!output) {
		cerr << "Could not allocate bitmap!" << endl;
		exit(1);
	}
	cout << "Blending images..." << endl;
	
	RGBQUAD color1;
	RGBQUAD color2;
	RGBQUAD color3;
	
	for (int x = 0; x < width; x++) {
		for (int y = 0; y < height; y++) {
			FreeImage_GetPixelColor(img1, x, y, &color1);
			FreeImage_GetPixelColor(img2, x, y, &color2);
			
			color3.rgbRed   = (char) ((1-t)*((double) color1.rgbRed  ) + (t)*((double) color2.rgbRed  ));	
			color3.rgbGreen = (char) ((1-t)*((double) color1.rgbGreen) + (t)*((double) color2.rgbGreen));
			color3.rgbBlue  = (char) ((1-t)*((double) color1.rgbBlue ) + (t)*((double) color2.rgbBlue ));
	
			FreeImage_SetPixelColor(output, x, y, &color3);
		}
	}
	return output;
}

/*
Morph img1 into img2 
*/
FIBITMAP* MorphImages(FIBITMAP *img1, 
	                  FIBITMAP *img2, 
	                  vector<FeatureVector> featuresImg1, 
	                  vector<FeatureVector> featuresImg2, 
	                  double t,
	                  double a, double b, double p) {
	
	int width = FreeImage_GetWidth(img1);
	int height = FreeImage_GetHeight(img1);
	flipFeatureVectors(featuresImg1, width, height);
	flipFeatureVectors(featuresImg2, width, height);
	
	if (DEBUG) {
		int rep = max(width, height) / 100;
		drawGrid(img1, rep);
		drawGrid(img2, rep);
	}
	
	FIBITMAP* d1 = distortImage(featuresImg1, featuresImg2, img1, t, a, b, p);
	FIBITMAP* d2 = distortImage(featuresImg2, featuresImg1, img2, (1-t), a, b, p);
	
	if (DEBUG) {
		FreeImage_Save(FIF_PNG, d1, "d1.png");
		FreeImage_Save(FIF_PNG, d2, "d2.png");
	}
	return blendImages(d1, d2, t);
	
}

///////////////////////////////////////////////////////////////////////////////
//
//  Driver functions follow. These should not be modified.
//
///////////////////////////////////////////////////////////////////////////////


void MorphDriver(char* img1_fn, char* img2_fn, char* linepairs, double t, char* outputFileName, double a, double b, double p) {
	FIBITMAP *img1 = GenericLoader(img1_fn, JPEG_ACCURATE);
	FIBITMAP *img2 = GenericLoader(img2_fn, JPEG_ACCURATE);
	
	if (!img1 || !img2) {
		cerr << "Could not load one of your input images." << endl;
		exit(1);
	}
	
	if (FreeImage_GetWidth(img1) != FreeImage_GetWidth(img2) || 
	    FreeImage_GetHeight(img1) != FreeImage_GetHeight(img2)) {	
		cout << "Both input images needs to be the same resolution." << endl;
		exit(1);
	} else {
		vector<FeatureVector> featuresImg1;
		vector<FeatureVector> featuresImg2;
		int numFeatures = readFeatureVectorsFromFile(linepairs, featuresImg1, featuresImg2);
		if (numFeatures < 0) {
			cerr << "Could not read features file. Exiting..." << endl;
			exit(1);
		} else {
			cout << "Read " << numFeatures << " features" << endl;
			FIBITMAP* result = MorphImages(img1, img2, featuresImg1, featuresImg2, t, a, b, p);
			if ((!result) || !FreeImage_Save(FIF_PNG, result, outputFileName)) {
				cerr << "Result could not be saved." << endl;
			}
		}
	}
	
	FreeImage_Unload(img1);
	FreeImage_Unload(img2);
}

int main(int argc, char* argv[]) {
		
	if (argc != 6 && argc != 9) {
		cerr << "Expected 5 arguments: image1 image2 line_pairs_file time[0..1] output.png [a] [b] [p]" << endl;
		exit(1);
	}
	
	char* img1      = argv[1];
	char* img2      = argv[2];
	char* linepairs = argv[3];
	double t        = atof(argv[4]);
	char* filename  = argv[5];
	
	if (t > 1.0 || t < 0.0) {
		cout << "Time t was out of range. Clamping to [0..1]" << endl;
		if (t > 1.0)
			t = 1.0;
		else
			t = 0.0;
	}
	

	double a = 0.5;
	double b = 1;
	double p = 0.2;	
	if (argc == 9) {
		a = atof(argv[6]);
		b = atof(argv[7]);
		p = atof(argv[8]);
	} 

	
	cout << "Morph Engine Loaded: ";
	cout << "Morping " << img1 << " into " << img2 << " at time t=" << t << ", generating " << filename << endl;
	cout << "Using parameters { a : " << a << ", b : " << b << ", p : " << p << " }" << endl;

	FreeImage_Initialise();	
	MorphDriver(img1, img2, linepairs, t, filename, a, b, p);
	FreeImage_DeInitialise();
	
	cout << endl;

	return 0;
}